#!/usr/bin/env python
"""
__init__.py. Includes wrapper for running dyPolyChord with different
dynamic goals.
"""
from dyPolyChord._version import __version__
from dyPolyChord.run_dynamic_ns import run_dypolychord
